/* 1541control - way overdesigned LCD addon for a 1541
   Copyright (C) 2007  Ingo Korb <ingo@akana.de>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License only.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   
   lcd.c: HD44780-compatible display access

*/

#include "config.h"
#include <avr/io.h>
#include <avr/pgmspace.h>
#include <util/delay.h>
#include <stdio.h>
#include "lcd.h"

#define DISPLAY_ON   0x0c

#define SET_DD_ADDRESS 0x80

uint8_t cursor_x;
uint8_t cursor_y;
uint8_t cursor_mode;

#ifdef USE_LCD_STDOUT
static FILE lcd_stdout = FDEV_SETUP_STREAM(lcd_putc, NULL, _FDEV_SETUP_WRITE);
#endif

static void inline lcd_strobe(void) {
  PORTA |=  _BV(PA6);
  PORTA &= ~_BV(PA6);
}

static void lcd_wait() {
  uint8_t tmp;

  PORTA = _BV(PA5);
  DDRA  = 0xf0;
  do {
    PORTA |= _BV(PA6);
    _delay_loop_1(5);
    tmp = PINA & 0x08;
    PORTA &= ~_BV(PA6);
    _delay_loop_1(5);
    PORTA |= _BV(PA6);
    _delay_loop_1(5);
    PORTA &= ~_BV(PA6);
  } while (tmp);
  PORTA = 0;
  DDRA = 0xff;
}

static void lcd_write(uint8_t rs, uint8_t value) {
  uint8_t tmp;

  if (rs)
    tmp = _BV(PA4);
  else
    tmp = 0;
  
  lcd_wait();

  PORTA = tmp | (value >> 4);
  lcd_strobe();
  PORTA = tmp | (value & 0x0f);
  lcd_strobe();
}

void lcd_init(void) {
  DDRA  = 0xff;
  PORTA = 0;
  _delay_ms(15);
  PORTA = 0x03;

  lcd_strobe();
  _delay_ms(5);
  lcd_strobe();
  _delay_us(100);
  lcd_strobe();
  lcd_wait();

  PORTA = 0x02;
  lcd_strobe();
  lcd_wait();

  lcd_write(0,0x28); // System Set: 4 Bit, 2 Zeilen, 5x7
  lcd_write(0,0x08); // Display aus
  lcd_write(0,0x0d); // Display an, Blinkcursor
  lcd_write(0,0x06); // Cursor beim Schreiben vorwaerts
  lcd_write(0,0x01); // Anzeige loeschen

  cursor_x = 0;
  cursor_y = 0;
  cursor_mode = LCD_CURSOR_BLOCK;

#ifdef USE_LCD_STDOUT
  stdout = &lcd_stdout;
#endif
}

static void updatecursor(void) {
  lcd_write(0,DISPLAY_ON | cursor_mode);
}

void lcd_clrscr(void) {
  lcd_write(0,0x01);
  cursor_x = 0;
  cursor_y = 0;
  updatecursor();
}

void lcd_setcursormode(uint8_t mode) {
  cursor_mode = mode;
  lcd_gotoxy(cursor_x,cursor_y);
}

void lcd_gotoxy(uint8_t x, uint8_t y) {
  cursor_x = x;
  cursor_y = y;
  updatecursor();
  lcd_write(0,SET_DD_ADDRESS + x + 0x40*y);
}

void lcd_putch(char c) {
  lcd_write(1,c);
  if (++cursor_x > LCD_COLUMNS-1)
    lcd_gotoxy(0,(cursor_y+1) & (LCD_ROWS-1));
}

#ifdef USE_LCD_STDOUT
int lcd_putc(char c, FILE *stream) {
  if (c == '\n') {
    lcd_gotoxy(0,(cursor_y+1) & (LCD_ROWS-1));
  } else {
    lcd_putch(c);
  }
  return 0;
}
#endif

void lcd_puts_P(prog_char *text) {
  uint8_t ch;

  while ((ch = pgm_read_byte(text++))) {
    lcd_putch(ch);
  }
  updatecursor();
}  

void lcd_putxy_P(uint8_t xpos, uint8_t ypos, prog_char *text) {
  if (xpos > (LCD_COLUMNS-1) || ypos > (LCD_ROWS-1)) return;

  lcd_gotoxy(xpos,ypos);
  lcd_puts_P(text);
}

void lcd_puts(char *text) {
  while (*text) {
    lcd_putch(*text++);
  }
  updatecursor();
}  

void lcd_putxy(uint8_t xpos, uint8_t ypos, char *text) {
  if (xpos > (LCD_COLUMNS-1) || ypos > (LCD_ROWS-1)) return;

  lcd_gotoxy(xpos,ypos);
  lcd_puts(text);
}

void lcd_customchar(uint8_t num, uint8_t b0, uint8_t b1,
		    uint8_t b2,  uint8_t b3, uint8_t b4,
		    uint8_t b5,  uint8_t b6, uint8_t b7) {
  if (num > 7) return;
  lcd_write(0,0x40+8*num);
  lcd_write(1,b0);
  lcd_write(1,b1);
  lcd_write(1,b2);
  lcd_write(1,b3);
  lcd_write(1,b4);
  lcd_write(1,b5);
  lcd_write(1,b6);
  lcd_write(1,b7);
  lcd_gotoxy(cursor_x,cursor_y);
}
