/* 1541control - way overdesigned LCD addon for a 1541
   Copyright (C) 2007  Ingo Korb <ingo@akana.de>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; version 2 of the License only.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   
   main.c: The glue that holds everything together

*/

#include "config.h"
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <util/delay.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "lcd.h"
#include "menu.h"
#include "uart.h"
#include "encoder.h"
#include "tracker.h"

static const prog_char dosmenu[] =
  "CBM-DOS\0"
  "Jiffy-DOS\0"
  "Dolphin-DOS\0"
  "64'er-DOS V3\0"
  "SpeedDOS\0"
  "S-JiffyDOS\0"
  "CBM-DOS ohne RAM\0"
  "Jiffy-DOS ohne RAM\0"
  "RAM-Dos\0"
  "Ger\341teadresse ~";

static const prog_char dosnames[] = {
 //1234567890123456
  "CBM-DOS      RAM"
  "Jiffy-DOS    RAM"
  "Dolphin-DOS  RAM"
  "64'er-DOS V3 RAM"
  "SpeedDOS     RAM"
  "S-JiffyDOS      "
  "CBM-DOS         "
  "Jiffy-DOS       "
  "RAM-DOS      RAM"
};

uint8_t *eeptr;
#define MIN_EEPROM_POS ((uint8_t *)10)
#define MAX_EEPROM_POS ((uint8_t *)508)

static void store_config(uint8_t dosnum, uint8_t devnum) {
  uint8_t *ptr;
  uint8_t data;

  data = dosnum + (devnum << 4);

  cli();
  if (eeptr >= MAX_EEPROM_POS) {
    eeptr = MIN_EEPROM_POS;
    eeprom_write_byte(eeptr, data);
    for (ptr = MIN_EEPROM_POS+1; ptr < MAX_EEPROM_POS; ptr++)
      eeprom_write_byte(ptr,0xff);
  } else {
    eeprom_write_byte(eeptr++, data);
  }
  EEAR = 0;
  sei();
}

static void reset_floppy() {
  DDRB |= _BV(PB3);
  _delay_ms(250);
  DDRB &= ~_BV(PB3);
}

static void set_device(const uint8_t num) {
  PORTD = (PORTD & ~(_BV(PD6) | _BV(PD7))) | ((num & 3) << PD6);
  reset_floppy();
}

static void set_dos(const uint8_t num) {
  PORTC = (PORTC & 0xf0) | (num & 0x0f);
  reset_floppy();
}

static char buffer[256];

static uint8_t dosnum,devnum;

static void exec_menu(void) {
  int i,j;
  uint8_t result;
  char *menuitems[10];

  lcd_clrscr();
  
  i = 0;
  j = 0;
  memcpy_P(buffer,dosmenu,sizeof(dosmenu));
  menuitems[0] = buffer;
  while (j < sizeof(dosmenu)) {
    if (buffer[j] == 0) {
      menuitems[++i] = buffer+j+1;
    }
    j++;
  }
  
  result = display_menu(menuitems,i,dosnum);
  
  if (result == i-1) {
    memset(buffer,0,sizeof(buffer));
    j = 0;
    for (i=8;i<12;i++) {
      menuitems[i-8] = buffer+j;
      j += sprintf_P(buffer+j,PSTR("%2d"),i)+1;
    }
    
    result = display_menu(menuitems,4,devnum);

    if (result != devnum) {
      devnum = result;
      set_device(devnum);
      store_config(dosnum,devnum);
    }
  } else {
    if (result != dosnum) {
      dosnum = result;
      set_dos(dosnum);
      store_config(dosnum,devnum);
    }
  }
}

static void update_track(void) {
  cli();
  if (tracker_flags & TRACK_CHANGED) {
    lcd_gotoxy(10,1);
    if (tracker_flags & TRACK_VALID) {
      printf_P(PSTR("T%02d%c"),current_track >> 1,(current_track & 1)?'h':' ');
    } else {
      lcd_puts_P(PSTR("T?? "));
    }
  }

  if (tracker_flags & DENSITY_CHANGED) {
    lcd_gotoxy(14,1);
    printf_P(PSTR("D%d"),current_density);
  }
  tracker_flags &= TRACK_VALID;
  sei();
}

static void show_idlescreen(void) {
  lcd_clrscr();
  memcpy_P(buffer,dosnames+16*dosnum,16);
  buffer[16] = 0;
  lcd_putxy(0,0,buffer);
  printf_P(PSTR("#%d"),devnum+8);
  cli();
  tracker_flags |= TRACK_CHANGED | DENSITY_CHANGED;
  update_track();
}



int main(void) {
  int8_t  prevencoder;
  uint8_t trackerreset,curkey;

  // Floppy im Reset halten
  PORTB &= ~_BV(PB3);
  DDRB  = _BV(PB3);

  init_serial();
  lcd_init();
  init_encoder();
  init_menu();

  // Restliche Ports initialisieren
  DDRC = 0x0f;                 // Select-Ausgaenge
  DDRD |= _BV(PD6) | _BV(PD7); // Gerteadresse

  // Letzte Konfiguration im EEPROM suchen
  eeptr = MIN_EEPROM_POS;
  while ((devnum = eeprom_read_byte(eeptr)) != 0xff) {
    dosnum = devnum;
    eeptr++;
    if (eeptr >= MAX_EEPROM_POS) break;
  }

  if (eeptr == 0) {
    dosnum = 0;
    devnum = 0;
  } else {
    devnum = (dosnum >> 4) & 3;
    dosnum &= 0x0f;
    if (dosnum >= sizeof(dosnames)/16) dosnum = 0;
  }

  lcd_clrscr();
  lcd_setcursormode(LCD_CURSOR_NONE);

  show_idlescreen();

  // Floppyreset wird hier als Seiteneffekt beendet
  set_device(devnum);
  set_dos(dosnum);

  prevencoder  = encoder_position;
  trackerreset = 0;

  sei();
  while (1) {
    curkey = key_state;

    // Bei Bewegung Menu aufrufen
    if (encoder_position != prevencoder) {
      exec_menu();
      prevencoder = encoder_position;
      show_idlescreen();
    }

    // Langer Druck: Floppyreset
    if (curkey && time_after(jiffies,last_key_change+1000)) {
      DDRB |= _BV(PB3);
      trackerreset = 2;
    }

    // Kurzer Druck: Tracker auf 18 stellen
    if (key_state && !trackerreset)
      trackerreset = 1;

    if (!key_state && trackerreset) {
      if (trackerreset == 1) {
	cli();
	current_track = 18<<1;
	tracker_flags = TRACK_VALID | TRACK_CHANGED;
	sei();
      } else if (trackerreset == 2) {
	DDRB &= ~_BV(PB3);
      }
      trackerreset = 0;
    }

    // Trackanzeige aktualisieren
    if (tracker_flags)
      update_track();
  }

  return 0;
}
